@@ -71,6 +71,16 @@ struct raft_buffer
     size_t len; /* Length of the buffer. */
 };
 
+/**
+ * A type for storing unknown bools.
+ */
+typedef enum {
+    raft_tribool_unknown,
+    raft_tribool_true,
+    raft_tribool_false,
+} raft_tribool;
+#define TO_RAFT_TRIBOOL(b)  ((b) ? raft_tribool_true : raft_tribool_false)
+
 /**
  * Server role codes.
  */
@@ -252,8 +262,9 @@ struct raft_request_vote
  */
 struct raft_request_vote_result
 {
-    raft_term term;    /* Receiver's current term (candidate updates itself). */
-    bool vote_granted; /* True means candidate received vote. */
+    raft_term term;        /* Receiver's current term (candidate updates itself). */
+    bool vote_granted;     /* True means candidate received vote. */
+    raft_tribool pre_vote; /* The response to a pre-vote RequestVote or not. */
 };
 
 /**
@@ -416,6 +427,11 @@ struct raft_tracer
      */
     void *impl;
 
+    /**
+     * Whether this tracer should emit messages.
+     */
+    bool enabled;
+
     /**
      * Emit the given trace message, possibly decorating it with the provided
      * metadata.
@@ -508,12 +524,13 @@ enum { RAFT_UNAVAILABLE, RAFT_FOLLOWER, RAFT_CANDIDATE, RAFT_LEADER };
  */
 struct raft_progress
 {
-    unsigned short state;      /* Probe, pipeline or snapshot. */
-    raft_index next_index;     /* Next entry to send. */
-    raft_index match_index;    /* Highest index reported as replicated. */
-    raft_index snapshot_index; /* Last index of most recent snapshot sent. */
-    raft_time last_send;       /* Timestamp of last AppendEntries RPC. */
-    bool recent_recv;          /* A msg was received within election timeout. */
+    unsigned short state;         /* Probe, pipeline or snapshot. */
+    raft_index next_index;        /* Next entry to send. */
+    raft_index match_index;       /* Highest index reported as replicated. */
+    raft_index snapshot_index;    /* Last index of most recent snapshot sent. */
+    raft_time last_send;          /* Timestamp of last AppendEntries RPC. */
+    raft_time snapshot_last_send; /* Timestamp of last InstallSnaphot RPC. */
+    bool recent_recv;             /* A msg was received within election timeout. */
 };
 
 struct raft; /* Forward declaration. */
@@ -560,16 +577,16 @@ struct raft
      *
      * The possible scenarios are:
      *
-     * 1. #configuration_index and #configuration_uncommited_index are both
+     * 1. #configuration_index and #configuration_uncommitted_index are both
      *    zero. This should only happen when a brand new server starts joining a
      *    cluster and is waiting to receive log entries from the current
      *    leader. In this case #configuration must be empty and have no servers.
      *
-     * 2. #configuration_index is non-zero while #configuration_uncommited_index
+     * 2. #configuration_index is non-zero while #configuration_uncommitted_index
      *    is zero. In this case the content of #configuration must match the one
      *    of the log entry at #configuration_index.
      *
-     * 3. #configuration_index and #configuration_uncommited_index are both
+     * 3. #configuration_index and #configuration_uncommitted_index are both
      *    non-zero, with the latter being greater than the former. In this case
      *    the content of #configuration must match the one of the log entry at
      *    #configuration_uncommitted_index.
@@ -612,6 +629,12 @@ struct raft
      */
     unsigned heartbeat_timeout;
 
+    /*
+     * When the leader sends an InstallSnapshot RPC to a follower it will consider
+     * the RPC as failed after this timeout and retry.
+     */
+    unsigned install_snapshot_timeout;
+
     /*
      * The fields below hold the part of the server's volatile state which is
      * always applicable regardless of the whether the server is follower,
@@ -767,6 +790,11 @@ RAFT_API void raft_set_election_timeout(struct raft *r, unsigned msecs);
  */
 RAFT_API void raft_set_heartbeat_timeout(struct raft *r, unsigned msecs);
 
+/**
+ * Set the snapshot install timeout.
+ */
+RAFT_API void raft_set_install_snapshot_timeout(struct raft *r, unsigned msecs);
+
 /**
  * Number of outstanding log entries before starting a new snapshot. The default
  * is 1024.
@@ -800,7 +828,7 @@ RAFT_API void raft_set_max_catch_up_round_duration(struct raft *r,
                                                    unsigned msecs);
 
 /**
- * Return a human-readable description of the last error occured.
+ * Return a human-readable description of the last error occurred.
  */
 RAFT_API const char *raft_errmsg(struct raft *r);
 
