@@ -13,21 +13,17 @@
 #include "snapshot.h"
 #include "tracing.h"
 
-/* Set to 1 to enable tracing. */
-#if 0
 #define tracef(...) Tracef(r->tracer, __VA_ARGS__)
-#else
-#define tracef(...)
-#endif
 
 /* Defaults */
 #define HEARTBEAT_TIMEOUT 100
+#define INSTALL_SNAPSHOT_TIMEOUT 30000
 #define ELECTION_TIMEOUT 1000
 #define NETWORK_LATENCY 15
 #define DISK_LATENCY 10
 
 /* To keep in sync with raft.h */
-#define N_MESSAGE_TYPES 5
+#define N_MESSAGE_TYPES 6
 
 /* Maximum number of peer stub instances connected to a certain stub
  * instance. This should be enough for testing purposes. */
@@ -367,7 +363,7 @@ out:
     raft_free(send);
 }
 
-/* Release the memory used by the given message tramsmit object. */
+/* Release the memory used by the given message transmit object. */
 static void ioDestroyTransmit(struct transmit *transmit)
 {
     struct raft_message *message;
@@ -925,9 +921,11 @@ static int serverInit(struct raft_fixture *f, unsigned i, struct raft_fsm *fsm)
     }
     raft_set_election_timeout(&s->raft, ELECTION_TIMEOUT);
     raft_set_heartbeat_timeout(&s->raft, HEARTBEAT_TIMEOUT);
+    raft_set_install_snapshot_timeout(&s->raft, INSTALL_SNAPSHOT_TIMEOUT);
     s->tracer.impl = (void *)&s->id;
     s->tracer.emit = emit;
     s->raft.tracer = &s->tracer;
+    raft_tracer_maybe_enable(&s->tracer, true);
     return 0;
 }
 
@@ -1264,6 +1262,7 @@ static void copyLeaderLog(struct raft_fixture *f)
         struct raft_buffer buf;
         buf.len = entry->buf.len;
         buf.base = raft_malloc(buf.len);
+        assert(buf.base != NULL);
         memcpy(buf.base, entry->buf.base, buf.len);
         rv = logAppend(&f->log, entry->term, entry->type, &buf, NULL);
         assert(rv == 0);
