@@ -3,6 +3,7 @@
 #include "assert.h"
 #include "configuration.h"
 #include "convert.h"
+#include "entry.h"
 #ifdef __GLIBC__
 #include "error.h"
 #endif
@@ -17,12 +18,7 @@
 #include "snapshot.h"
 #include "tracing.h"
 
-/* Set to 1 to enable tracing. */
-#if 0
 #define tracef(...) Tracef(r->tracer, __VA_ARGS__)
-#else
-#define tracef(...)
-#endif
 
 #ifndef max
 #define max(a, b) ((a) < (b) ? (b) : (a))
@@ -53,22 +49,10 @@ static void sendAppendEntriesCb(struct raft_io_send *send, const int status)
 
     if (r->state == RAFT_LEADER && i < r->configuration.n) {
         if (status != 0) {
-            tracef("failed to send append entries to server %u: %s",
+            tracef("failed to send append entries to server %llu: %s",
                    req->server_id, raft_strerror(status));
             /* Go back to probe mode. */
             progressToProbe(r, i);
-        } else {
-            /* Update the last_send timestamp: for a heartbeat_timeout
-             * milliseconds we'll be good and we won't need to contact followers
-             * again, since this was not an idle period.
-             *
-             * From Figure 3.1:
-             *
-             *   [Rules for Servers] Leaders: Upon election: send initial empty
-             *   AppendEntries RPCs (heartbeat) to each server; repeat during
-             * idle periods to prevent election timeouts
-             */
-            progressUpdateLastSend(r, i);
         }
     }
 
@@ -111,7 +95,7 @@ static int sendAppendEntries(struct raft *r,
      */
     args->leader_commit = r->commit_index;
 
-    tracef("send %u entries starting at %llu to server %u (last index %llu)",
+    tracef("send %u entries starting at %llu to server %llu (last index %llu)",
            args->n_entries, args->prev_log_index, server->id,
            logLastIndex(&r->log));
 
@@ -137,10 +121,11 @@ static int sendAppendEntries(struct raft *r,
     }
 
     if (progressState(r, i) == PROGRESS__PIPELINE) {
-        /* Optimitiscally update progress. */
+        /* Optimistically update progress. */
         progressOptimisticNextIndex(r, i, req->index + req->n);
     }
 
+    progressUpdateLastSend(r, i);
     return 0;
 
 err_after_req_alloc:
@@ -237,7 +222,7 @@ static void sendSnapshotGetCb(struct raft_io_snapshot_get *get,
     req->snapshot = snapshot;
     req->send.data = req;
 
-    tracef("sending snapshot with last index %llu to %u", snapshot->index,
+    tracef("sending snapshot with last index %llu to %llu", snapshot->index,
            server->id);
 
     rv = r->io->send(r->io, &req->send, &message, sendInstallSnapshotCb);
@@ -286,6 +271,7 @@ static int sendSnapshot(struct raft *r, const unsigned i)
         goto err_after_req_alloc;
     }
 
+    progressUpdateSnapshotLastSend(r, i);
     return 0;
 
 err_after_req_alloc:
@@ -299,6 +285,7 @@ err:
 int replicationProgress(struct raft *r, unsigned i)
 {
     struct raft_server *server = &r->configuration.servers[i];
+    bool progress_state_is_snapshot = progressState(r, i) == PROGRESS__SNAPSHOT;
     raft_index snapshot_index = logSnapshotIndex(&r->log);
     raft_index next_index = progressNextIndex(r, i);
     raft_index prev_index;
@@ -328,8 +315,8 @@ int replicationProgress(struct raft *r, unsigned i)
     if (next_index == 1) {
         /* We're including the very first entry, so prevIndex and prevTerm are
          * null. If the first entry is not available anymore, send the last
-         * snapshot. */
-        if (snapshot_index > 0) {
+         * snapshot if we're not already sending one. */
+        if (snapshot_index > 0 && !progress_state_is_snapshot) {
             raft_index last_index = logLastIndex(&r->log);
             assert(last_index > 0); /* The log can't be empty */
             goto send_snapshot;
@@ -341,18 +328,33 @@ int replicationProgress(struct raft *r, unsigned i)
          * next_index - 1. */
         prev_index = next_index - 1;
         prev_term = logTermOf(&r->log, prev_index);
-        /* If the entry is not anymore in our log, send the last snapshot. */
-        if (prev_term == 0) {
+        /* If the entry is not anymore in our log, send the last snapshot if we're
+         * not doing so already. */
+        if (prev_term == 0 && !progress_state_is_snapshot) {
             assert(prev_index < snapshot_index);
             tracef("missing entry at index %lld -> send snapshot", prev_index);
             goto send_snapshot;
         }
     }
 
+    /* Send empty AppendEntries RPC when installing a snaphot */
+    if (progress_state_is_snapshot) {
+        prev_index = logLastIndex(&r->log);
+        prev_term = logLastTerm(&r->log);
+    }
+
     return sendAppendEntries(r, i, prev_index, prev_term);
 
 send_snapshot:
-    return sendSnapshot(r, i);
+    if (progressGetRecentRecv(r, i)) {
+        /* Only send a snapshot when we have heard from the server */
+        return sendSnapshot(r, i);
+    } else {
+        /* Send empty AppendEntries RPC when we haven't heard from the server */
+        prev_index = logLastIndex(&r->log);
+        prev_term = logLastTerm(&r->log);
+        return sendAppendEntries(r, i, prev_index, prev_term);
+    }
 }
 
 /* Possibly trigger I/O requests for newly appended log entries or heartbeats.
@@ -381,7 +383,7 @@ static int triggerAll(struct raft *r)
         rv = replicationProgress(r, i);
         if (rv != 0 && rv != RAFT_NOCONNECTION) {
             /* This is not a critical failure, let's just log it. */
-            tracef("failed to send append entries to server %u: %s (%d)",
+            tracef("failed to send append entries to server %llu: %s (%d)",
                    server->id, raft_strerror(rv), rv);
         }
     }
@@ -494,9 +496,6 @@ static void appendLeaderCb(struct raft_io_append *req, int status)
         goto out;
     }
 
-    /* If Check if we have reached a quorum. */
-    server_index = configurationIndexOf(&r->configuration, r->id);
-
     /* Only update the next index if we are part of the current
      * configuration. The only case where this is not true is when we were
      * asked to remove ourselves from the cluster.
@@ -507,11 +506,9 @@ static void appendLeaderCb(struct raft_io_append *req, int status)
      *   leader can manage a cluster that does not include itself; it
      *   replicates log entries but does not count itself in majorities.
      */
+    server_index = configurationIndexOf(&r->configuration, r->id);
     if (server_index < r->configuration.n) {
         r->leader_state.progress[server_index].match_index = r->last_stored;
-    } else {
-        const struct raft_entry *entry = logGet(&r->log, r->last_stored);
-        assert(entry->type == RAFT_CHANGE);
     }
 
     /* Check if we can commit some new entries. */
@@ -551,7 +548,13 @@ static int appendLeader(struct raft *r, raft_index index)
 
     /* We expect this function to be called only when there are actually
      * some entries to write. */
-    assert(n > 0);
+    if (n == 0) {
+        assert(false);
+        tracef("No log entries found at index %llu", index);
+        ErrMsgPrintf(r->errmsg, "No log entries found at index %llu", index);
+        rv = RAFT_SHUTDOWN;
+        goto err_after_entries_acquired;
+    }
 
     /* Allocate a new request. */
     request = raft_malloc(sizeof *request);
@@ -686,7 +689,7 @@ int replicationUpdate(struct raft *r,
                                        result->last_log_index);
         if (retry) {
             /* Retry, ignoring errors. */
-            tracef("log mismatch -> send old entries to %u", server->id);
+            tracef("log mismatch -> send old entries to %llu", server->id);
             replicationProgress(r, i);
         }
         return 0;
@@ -757,7 +760,7 @@ int replicationUpdate(struct raft *r,
     i = configurationIndexOf(&r->configuration, server->id);
 
     if (i < r->configuration.n) {
-        /* If we are transfering leadership to this follower, check if its log
+        /* If we are transferring leadership to this follower, check if its log
          * is now up-to-date and, if so, send it a TimeoutNow RPC (unless we
          * already did). */
         if (r->transfer != NULL && r->transfer->id == server->id) {
@@ -849,6 +852,12 @@ static void appendFollowerCb(struct raft_io_append *req, int status)
         goto out;
     }
 
+    /* We received an InstallSnapshot RCP while these entries were being
+     * persisted to disk */
+    if (replicationInstallSnapshotBusy(r)) {
+        goto out;
+    }
+
     i = updateLastStored(r, request->index, args->entries, args->n_entries);
 
     /* If none of the entries that we persisted is present anymore in our
@@ -982,7 +991,6 @@ static int deleteConflictingEntries(struct raft *r,
             if (entry_index <= r->commit_index) {
                 /* Should never happen; something is seriously wrong! */
                 tracef("new index conflicts with committed entry -> shutdown");
-
                 return RAFT_SHUTDOWN;
             }
 
@@ -1065,7 +1073,9 @@ int replicationAppend(struct raft *r,
     n = args->n_entries - i; /* Number of new entries */
 
     /* If this is an empty AppendEntries, there's nothing to write. However we
-     * still want to check if we can commit some entry.
+     * still want to check if we can commit some entry. However, don't commit
+     * anything while a snapshot install is busy, r->last_stored will be 0 in
+     * that case.
      *
      * From Figure 3.1:
      *
@@ -1074,8 +1084,9 @@ int replicationAppend(struct raft *r,
      *   entry).
      */
     if (n == 0) {
-        if (args->leader_commit > r->commit_index) {
-            r->commit_index = min(args->leader_commit, logLastIndex(&r->log));
+        if ((args->leader_commit > r->commit_index)
+             && !replicationInstallSnapshotBusy(r)) {
+            r->commit_index = min(args->leader_commit, r->last_stored);
             rv = replicationApply(r);
             if (rv != 0) {
                 return rv;
@@ -1095,6 +1106,7 @@ int replicationAppend(struct raft *r,
 
     request->raft = r;
     request->args = *args;
+    /* Index of first new entry */
     request->index = args->prev_log_index + 1 + i;
 
     /* Update our in-memory log to reflect that we received these entries. We'll
@@ -1102,11 +1114,19 @@ int replicationAppend(struct raft *r,
      * that we issue below actually completes.  */
     for (j = 0; j < n; j++) {
         struct raft_entry *entry = &args->entries[i + j];
+         /* TODO This copy should not strictly be necessary, as the batch logic will
+          * take care of freeing the batch buffer in which the entries are received.
+          * However, this would lead to memory spikes in certain edge cases.
+          * https://github.com/canonical/dqlite/issues/276
+          */
+        struct raft_entry copy = {0};
+        rv = entryCopy(entry, &copy);
+        if (rv != 0) {
+            goto err_after_request_alloc;
+        }
 
-        rv = logAppend(&r->log, entry->term, entry->type, &entry->buf,
-                       entry->batch);
+        rv = logAppend(&r->log, copy.term, copy.type, &copy.buf, NULL);
         if (rv != 0) {
-            /* TODO: we should revert any changes we made to the log */
             goto err_after_request_alloc;
         }
     }
@@ -1119,6 +1139,12 @@ int replicationAppend(struct raft *r,
     }
 
     assert(request->args.n_entries == n);
+    if (request->args.n_entries == 0) {
+        tracef("No log entries found at index %llu", request->index);
+        ErrMsgPrintf(r->errmsg, "No log entries found at index %llu", request->index);
+        rv = RAFT_SHUTDOWN;
+        goto err_after_acquire_entries;
+    }
 
     request->req.data = request;
     rv = r->io->append(r->io, &request->req, request->args.entries,
@@ -1128,15 +1154,22 @@ int replicationAppend(struct raft *r,
         goto err_after_acquire_entries;
     }
 
-    raft_free(args->entries);
-
+    entryBatchesDestroy(args->entries, args->n_entries);
     return 0;
 
 err_after_acquire_entries:
+    /* Release the entries related to the IO request */
     logRelease(&r->log, request->index, request->args.entries,
                request->args.n_entries);
 
 err_after_request_alloc:
+    /* Release all entries added to the in-memory log, making
+     * sure the in-memory log and disk don't diverge, leading
+     * to future log entries not being persisted to disk.
+     */
+    if (j != 0) {
+        logTruncate(&r->log, request->index);
+    }
     raft_free(request);
 
 err:
@@ -1226,15 +1259,17 @@ int replicationInstallSnapshot(struct raft *r,
     *async = false;
 
     /* If we are taking a snapshot ourselves or installing a snapshot, ignore
-     * the request, the leader will weventually retry. TODO: we should do
+     * the request, the leader will eventually retry. TODO: we should do
      * something smarter. */
     if (r->snapshot.pending.term != 0 || r->snapshot.put.data != NULL) {
         *async = true;
-        return 0;
+        tracef("already taking or installing snapshot");
+        return RAFT_BUSY;
     }
 
     /* If our last snapshot is more up-to-date, this is a no-op */
     if (r->log.snapshot.last_index >= args->last_index) {
+        tracef("have more recent snapshot");
         *rejected = 0;
         return 0;
     }
@@ -1242,13 +1277,14 @@ int replicationInstallSnapshot(struct raft *r,
     /* If we already have all entries in the snapshot, this is a no-op */
     local_term = logTermOf(&r->log, args->last_index);
     if (local_term != 0 && local_term >= args->last_term) {
+        tracef("have all entries");
         *rejected = 0;
         return 0;
     }
 
     *async = true;
 
-    /* Premptively update our in-memory state. */
+    /* Preemptively update our in-memory state. */
     logRestore(&r->log, args->last_index, args->last_term);
 
     r->last_stored = 0;
@@ -1280,6 +1316,7 @@ int replicationInstallSnapshot(struct raft *r,
                              0 /* zero trailing means replace everything */,
                              &r->snapshot.put, snapshot, installSnapshotCb);
     if (rv != 0) {
+        tracef("snapshot_put failed %d", rv);
         goto err_after_bufs_alloc;
     }
 
@@ -1355,7 +1392,8 @@ static void applyChange(struct raft *r, const raft_index index)
          *   steps down once the Cnew entry is committed.
          */
         server = configurationGet(&r->configuration, r->id);
-        if (server == NULL) {
+        if (server == NULL || server->role != RAFT_VOTER) {
+            tracef("leader removed from config or no longer voter server: %p", (void*)server);
             convertToFollower(r);
         }
 
@@ -1372,9 +1410,9 @@ static bool shouldTakeSnapshot(struct raft *r)
         return false;
     }
 
-    /* If a snapshot is already in progress, we don't want to start another
-     *  one. */
-    if (r->snapshot.pending.term != 0) {
+    /* If a snapshot is already in progress or we're installing a snapshot, we
+     * don't want to start another one. */
+    if (r->snapshot.pending.term != 0 || r->snapshot.put.data != NULL) {
         return false;
     };
 
@@ -1460,7 +1498,7 @@ abort:
 int replicationApply(struct raft *r)
 {
     raft_index index;
-    int rv;
+    int rv = 0;
 
     assert(r->state == RAFT_LEADER || r->state == RAFT_FOLLOWER);
     assert(r->last_applied <= r->commit_index);
@@ -1472,6 +1510,11 @@ int replicationApply(struct raft *r)
 
     for (index = r->last_applied + 1; index <= r->commit_index; index++) {
         const struct raft_entry *entry = logGet(&r->log, index);
+        if (entry == NULL) {
+            /* This can happen while installing a snapshot */
+            tracef("replicationApply - ENTRY NULL");
+            return 0;
+        }
 
         assert(entry->type == RAFT_COMMAND || entry->type == RAFT_BARRIER ||
                entry->type == RAFT_CHANGE);
@@ -1544,4 +1587,9 @@ void replicationQuorum(struct raft *r, const raft_index index)
     return;
 }
 
+inline bool replicationInstallSnapshotBusy(struct raft *r)
+{
+    return r->last_stored == 0 && r->snapshot.put.data != NULL;
+}
+
 #undef tracef
