@@ -19,18 +19,60 @@
 #include "uv_encoding.h"
 #include "uv_os.h"
 
-/* Set to 1 to enable tracing. */
-#if 0
-#define tracef(...) Tracef(c->uv->tracer, __VA_ARGS__)
-#else
-#define tracef(...)
-#endif
+#define tracef(...) Tracef(uv->tracer, __VA_ARGS__)
 
 /* Retry to connect to peer servers every second.
  *
  * TODO: implement an exponential backoff instead.  */
 #define CONNECT_RETRY_DELAY 1000
 
+/* Cleans up files that are no longer used by the system */
+static int uvMaintenance(const char *dir, char* errmsg)
+{
+    struct uv_fs_s req;
+    struct uv_dirent_s entry;
+    int n;
+    int i;
+    int rv;
+    int rv2;
+
+    n = uv_fs_scandir(NULL, &req, dir, 0, NULL);
+    if (n < 0) {
+        ErrMsgPrintf(errmsg, "scan data directory: %s", uv_strerror(n));
+        return RAFT_IOERR;
+    }
+
+    rv = 0;
+    for (i = 0; i < n; i++) {
+        const char *filename;
+        rv = uv_fs_scandir_next(&req, &entry);
+        assert(rv == 0); /* Can't fail in libuv */
+
+        filename = entry.name;
+        /* Remove leftover tmp-files */
+        if (strncmp(filename, TMP_FILE_PREFIX, strlen(TMP_FILE_PREFIX)) == 0) {
+            UvFsRemoveFile(dir, filename, errmsg); /* Ignore errors */
+            continue;
+        }
+
+        /* Remove orphaned snapshot files */
+        bool orphan = false;
+        if ((UvSnapshotIsOrphan(dir, filename, &orphan) == 0) && orphan) {
+            UvFsRemoveFile(dir, filename, errmsg); /* Ignore errors */
+            continue;
+        }
+
+        /* Remove orphaned snapshot metadata files */
+        if ((UvSnapshotMetaIsOrphan(dir, filename, &orphan) == 0) && orphan) {
+            UvFsRemoveFile(dir, filename, errmsg); /* Ignore errors */
+        }
+    }
+
+    rv2 = uv_fs_scandir_next(&req, &entry);
+    assert(rv2 == UV_EOF);
+    return rv;
+}
+
 /* Implementation of raft_io->config. */
 static int uvInit(struct raft_io *io, raft_id id, const char *address)
 {
@@ -46,7 +88,7 @@ static int uvInit(struct raft_io *io, raft_id id, const char *address)
         return rv;
     }
 
-    /* Proble file system capabilities */
+    /* Probe file system capabilities */
     rv = UvFsProbeCapabilities(uv->dir, &direct_io, &uv->async_io, io->errmsg);
     if (rv != 0) {
         return rv;
@@ -54,6 +96,11 @@ static int uvInit(struct raft_io *io, raft_id id, const char *address)
     uv->direct_io = direct_io != 0;
     uv->block_size = direct_io != 0 ? direct_io : 4096;
 
+    rv = uvMaintenance(uv->dir, io->errmsg);
+    if (rv != 0) {
+        return rv;
+    }
+
     rv = uvMetadataLoad(uv->dir, &metadata, io->errmsg);
     if (rv != 0) {
         return rv;
@@ -337,8 +384,12 @@ static int uvLoadSnapshotAndEntries(struct uv *uv,
         if (rv != 0) {
             goto err;
         }
-        if (segments != NULL && !segments[0].is_open) {
-            *start_index = segments[0].first_index;
+        if (segments != NULL) {
+            if (segments[0].is_open) {
+                *start_index = (*snapshot)->index + 1;
+            } else {
+                *start_index = segments[0].first_index;
+            }
         } else {
             *start_index = (*snapshot)->index + 1;
         }
@@ -538,7 +589,12 @@ static raft_time uvTime(struct raft_io *io)
 /* Implementation of raft_io->random. */
 static int uvRandom(struct raft_io *io, int min, int max)
 {
-    (void)io;
+    static bool initialized = false;
+    if (!initialized) {
+        struct uv *uv = io->impl;
+        srand((unsigned)uv_now(uv->loop) + (unsigned)uv->id);
+        initialized = true;
+    }
     return min + (abs(rand()) % (max - min));
 }
 
@@ -576,15 +632,21 @@ int raft_uv_init(struct raft_io *io,
 
     uv->io = io;
     uv->loop = loop;
-    strcpy(uv->dir, dir);
+    strncpy(uv->dir, dir, sizeof(uv->dir)-1);
+    uv->dir[sizeof(uv->dir)-1] = '\0';
     uv->transport = transport;
     uv->transport->data = NULL;
-    uv->tracer = &NoopTracer;
+    uv->tracer = &StderrTracer;
     uv->id = 0; /* Set by raft_io->config() */
     uv->state = UV__PRISTINE;
     uv->errored = false;
     uv->direct_io = false;
     uv->async_io = false;
+#ifdef LZ4_ENABLED
+    uv->snapshot_compression = true;
+#else
+    uv->snapshot_compression = false;
+#endif
     uv->segment_size = UV__MAX_SEGMENT_SIZE;
     uv->block_size = 0;
     QUEUE_INIT(&uv->clients);
@@ -661,6 +723,19 @@ void raft_uv_set_block_size(struct raft_io *io, size_t size)
     uv->block_size = size;
 }
 
+int raft_uv_set_snapshot_compression(struct raft_io *io, bool compressed)
+{
+    struct uv *uv;
+    uv = io->impl;
+#ifndef LZ4_AVAILABLE
+    if (compressed) {
+        return RAFT_INVALID;
+    }
+#endif
+    uv->snapshot_compression = compressed;
+    return 0;
+}
+
 void raft_uv_set_connect_retry_delay(struct raft_io *io, unsigned msecs)
 {
     struct uv *uv;
