@@ -270,6 +270,12 @@ out:
         if (rv != 0) {
             uv->errored = true;
         }
+    } else if (s->finalize) {
+        /* If there are no more append_pending_reqs, this segment
+         * must be finalized here in case we don't receive AppendEntries
+         * RPCs anymore (could happen during a Snapshot install, causing
+         * the BarrierCb to never fire) */
+        uvAliveSegmentFinalize(s);
     }
 }
 
@@ -699,6 +705,24 @@ static void uvFinalizeCurrentAliveSegmentOnceIdle(struct uv *uv)
     }
 }
 
+bool UvBarrierReady(struct uv *uv)
+{
+    if (uv->barrier == NULL) {
+        return true;
+    }
+
+    queue *head;
+    QUEUE_FOREACH(head, &uv->append_segments)
+    {
+        struct uvAliveSegment *segment;
+        segment = QUEUE_DATA(head, struct uvAliveSegment, queue);
+        if (segment->barrier == uv->barrier) {
+                return false;
+        }
+    }
+    return true;
+}
+
 int UvBarrier(struct uv *uv,
               raft_index next_index,
               struct UvBarrier *barrier,
