@@ -6,6 +6,7 @@
 #include <unistd.h>
 
 #include "assert.h"
+#include "compress.h"
 #include "err.h"
 #include "heap.h"
 #include "uv_os.h"
@@ -264,8 +265,58 @@ int UvFsMakeFile(const char *dir,
                  unsigned n_bufs,
                  char *errmsg)
 {
+    int rv;
+    char tmp_filename[UV__FILENAME_LEN+1] = {0};
+    char path[UV__PATH_SZ] = {0};
+    char tmp_path[UV__PATH_SZ] = {0};
+
+    /* Create a temp file with the given content
+     * TODO as of libuv 1.34.0, use `uv_fs_mkstemp` */
+    size_t sz = sizeof(tmp_filename);
+    rv = snprintf(tmp_filename, sz, TMP_FILE_FMT, filename);
+    if (rv < 0 || rv >= (int) sz) {
+        return rv;
+    }
     int flags = UV_FS_O_WRONLY | UV_FS_O_CREAT | UV_FS_O_EXCL;
-    return uvFsWriteFile(dir, filename, flags, bufs, n_bufs, errmsg);
+    rv = uvFsWriteFile(dir, tmp_filename, flags, bufs, n_bufs, errmsg);
+    if (rv != 0) {
+        goto err_after_tmp_create;
+    }
+
+    /* Check if the file exists */
+    bool exists = false;
+    rv = UvFsFileExists(dir, filename, &exists, errmsg);
+    if (rv != 0) {
+        goto err_after_tmp_create;
+    }
+    if (exists) {
+        rv = -1;
+        goto err_after_tmp_create;
+    }
+
+    /* Rename the temp file. Remark that there is a race between the existence
+     * check and the rename, there is no `renameat2` equivalent in libuv.
+     * However, in the current implementation this should pose no problems.*/
+    UvOsJoin(dir, tmp_filename, tmp_path);
+    UvOsJoin(dir, filename, path);
+    rv = UvOsRename(tmp_path, path);
+    if (rv != 0) {
+        UvOsErrMsg(errmsg, "rename", rv);
+        goto err_after_tmp_create;
+    }
+
+    rv = UvFsSyncDir(dir, errmsg);
+    if (rv != 0) {
+        char ignored[RAFT_ERRMSG_BUF_SIZE];
+        UvFsRemoveFile(dir, filename, ignored);
+        return rv;
+    }
+
+    return 0;
+
+err_after_tmp_create:
+    UvFsRemoveFile(dir, tmp_filename, errmsg);
+    return rv;
 }
 
 int UvFsMakeOrOverwriteFile(const char *dir,
@@ -339,69 +390,6 @@ err:
     return RAFT_IOERR;
 }
 
-int UvFsFileHasOnlyTrailingZeros(uv_file fd, bool *flag, char *errmsg)
-{
-    struct raft_buffer buf;
-    off_t size;
-    off_t offset;
-    size_t i;
-    int rv;
-
-    /* Save the current offset. */
-    offset = lseek(fd, 0, SEEK_CUR);
-
-    /* Figure the size of the rest of the file. */
-    size = lseek(fd, 0, SEEK_END);
-    if (size == -1) {
-        UvOsErrMsg(errmsg, "lseek", -errno);
-        return RAFT_IOERR;
-    }
-    size -= offset;
-
-    /* Reposition the file descriptor offset to the original offset. */
-    offset = lseek(fd, offset, SEEK_SET);
-    if (offset == -1) {
-        UvOsErrMsg(errmsg, "lseek", -errno);
-        return RAFT_IOERR;
-    }
-
-    buf.len = (size_t)size;
-    buf.base = HeapMalloc(buf.len);
-    if (buf.base == NULL) {
-        ErrMsgOom(errmsg);
-        return RAFT_NOMEM;
-    }
-
-    rv = UvFsReadInto(fd, &buf, errmsg);
-    if (rv != 0) {
-        return rv;
-    }
-
-    for (i = 0; i < (size_t)size; i++) {
-        if (((char *)buf.base)[i] != 0) {
-            *flag = false;
-            goto done;
-        }
-    }
-
-    *flag = true;
-
-done:
-    HeapFree(buf.base);
-
-    return 0;
-}
-
-bool UvFsIsAtEof(uv_file fd)
-{
-    off_t offset;
-    off_t size;
-    offset = lseek(fd, 0, SEEK_CUR); /* Get the current offset */
-    size = lseek(fd, 0, SEEK_END);   /* Get file size */
-    lseek(fd, offset, SEEK_SET);     /* Restore current offset */
-    return offset == size;           /* Compare current offset and size */
-}
-
 int UvFsReadInto(uv_file fd, struct raft_buffer *buf, char *errmsg)
 {
     int rv;
@@ -582,7 +570,7 @@ static int probeDirectIO(int fd, size_t *size, char *errmsg)
         switch (fs_info.f_type) {
             case 0x01021994: /* TMPFS_MAGIC */
             case 0x2fc12fc1: /* ZFS magic */
-            case 0x24051905: /* UBIFS Suprt magic */
+            case 0x24051905: /* UBIFS Support magic */
                 *size = 0;
                 return 0;
             default:
@@ -598,7 +586,7 @@ static int probeDirectIO(int fd, size_t *size, char *errmsg)
         }
     }
 
-    /* Try to peform direct I/O, using various buffer size. */
+    /* Try to perform direct I/O, using various buffer size. */
     *size = 4096;
     while (*size >= 512) {
         buf = raft_aligned_alloc(*size, *size);
@@ -696,6 +684,11 @@ static int probeAsyncIO(int fd, size_t size, bool *ok, char *errmsg)
     /* Fetch the response: will block until done. */
     n_events = UvOsIoGetevents(ctx, 1, 1, &event, NULL);
     assert(n_events == 1);
+    if (n_events != 1) {
+        /* UNTESTED */
+        UvOsErrMsg(errmsg, "UvOsIoGetevents", n_events);
+        return RAFT_IOERR;
+    }
 
     /* Release the write buffer. */
     raft_aligned_free(size, buf);
