@@ -9,14 +9,11 @@
 #include "byte.h"
 #include "configuration.h"
 #include "entry.h"
+#include "heap.h"
 #include "uv.h"
 #include "uv_encoding.h"
 
-#if 0
-#define tracef(...) Tracef(c->uv->tracer, __VA_ARGS__)
-#else
-#define tracef(...)
-#endif
+#define tracef(...) Tracef(uv->tracer, __VA_ARGS__)
 
 /* Check if the given filename matches the one of a closed segment (xxx-yyy), or
  * of an open segment (open-xxx), and fill the given info structure if so.
@@ -26,6 +23,7 @@ static bool uvSegmentInfoMatch(const char *filename, struct uvSegmentInfo *info)
 {
     int consumed;
     int matched;
+    size_t n;
     size_t filename_len = strnlen(filename, UV__FILENAME_LEN + 1);
 
     assert(filename_len < UV__FILENAME_LEN);
@@ -47,7 +45,9 @@ static bool uvSegmentInfoMatch(const char *filename, struct uvSegmentInfo *info)
     return false;
 
 match:
-    strcpy(info->filename, filename);
+    n = sizeof(info->filename) - 1;
+    strncpy(info->filename, filename, n);
+    info->filename[n] = '\0';
     return true;
 }
 
@@ -63,7 +63,7 @@ int uvSegmentInfoAppendIfMatch(const char *filename,
     /* Check if it's a closed or open filename */
     matched = uvSegmentInfoMatch(filename, &info);
 
-    /* If fhis is neither a closed or an open segment, return. */
+    /* If this is neither a closed or an open segment, return. */
     if (!matched) {
         *appended = false;
         return 0;
@@ -155,29 +155,47 @@ int uvSegmentKeepTrailing(struct uv *uv,
     return 0;
 }
 
-/* Open a segment file and read its format version. */
-static int uvOpenSegmentFile(struct uv *uv,
+/* Read a segment file and return its format version. */
+static int uvReadSegmentFile(struct uv *uv,
                              const char *filename,
-                             uv_file *fd,
+                             struct raft_buffer *buf,
                              uint64_t *format)
 {
     char errmsg[RAFT_ERRMSG_BUF_SIZE];
-    struct raft_buffer buf;
     int rv;
-    rv = UvFsOpenFileForReading(uv->dir, filename, fd, errmsg);
+    rv = UvFsReadFile(uv->dir, filename, buf, errmsg);
     if (rv != 0) {
-        ErrMsgTransfer(errmsg, uv->io->errmsg, "open file");
+        ErrMsgTransfer(errmsg, uv->io->errmsg, "read file");
         return RAFT_IOERR;
     }
-    buf.base = format;
-    buf.len = sizeof *format;
-    rv = UvFsReadInto(*fd, &buf, errmsg);
-    if (rv != 0) {
-        ErrMsgTransfer(errmsg, uv->io->errmsg, "read format");
-        UvOsClose(*fd);
+    if (buf->len < 8) {
+        ErrMsgPrintf(uv->io->errmsg, "file has only %zu bytes", buf->len);
+        HeapFree(buf->base);
         return RAFT_IOERR;
     }
-    *format = byteFlip64(*format);
+    *format = byteFlip64(*(uint64_t *)buf->base);
+    return 0;
+}
+
+/* Consume the content buffer, returning a pointer to the current position and
+ * advancing the offset of n bytes. Return an error if not enough bytes are
+ * available. */
+static int uvConsumeContent(const struct raft_buffer *content,
+                            size_t *offset,
+                            size_t n,
+                            void **data,
+                            char *errmsg)
+{
+    if (*offset + n > content->len) {
+        size_t remaining = content->len - *offset;
+        ErrMsgPrintf(errmsg, "short read: %zu bytes instead of %zu", remaining,
+                     n);
+        return RAFT_IOERR;
+    }
+    if (data != NULL) {
+        *data = &((uint8_t *)content->base)[*offset];
+    }
+    *offset += n;
     return 0;
 }
 
@@ -185,39 +203,45 @@ static int uvOpenSegmentFile(struct uv *uv,
  *
  * Set @last to #true if the loaded batch is the last one. */
 static int uvLoadEntriesBatch(struct uv *uv,
-                              const int fd,
+                              const struct raft_buffer *content,
                               struct raft_entry **entries,
                               unsigned *n_entries,
-                              off_t *offset, /* Offset of last batch */
+                              size_t *offset, /* Offset of last batch */
                               bool *last)
 {
-    uint64_t preamble[2];      /* CRC32 checksums and number of raft entries */
+    void *checksums;           /* CRC32 checksums */
+    void *batch;               /* Entries batch */
     unsigned long n;           /* Number of entries in the batch */
     unsigned max_n;            /* Maximum number of entries we expect */
     unsigned i;                /* Iterate through the entries */
-    struct raft_buffer buf;    /* Read buffer */
     struct raft_buffer header; /* Batch header */
     struct raft_buffer data;   /* Batch data */
     uint32_t crc1;             /* Target checksum */
     uint32_t crc2;             /* Actual checksum */
     char errmsg[RAFT_ERRMSG_BUF_SIZE];
+    size_t start;
     int rv;
 
     /* Save the current offset, to provide more information when logging. */
-    *offset = lseek(fd, 0, SEEK_CUR);
+    start = *offset;
 
-    /* Read the preamble, consisting of the checksums for the batch header and
-     * data buffers and the first 8 bytes of the header buffer, which contains
-     * the number of entries in the batch. */
-    buf.base = preamble;
-    buf.len = sizeof preamble;
-    rv = UvFsReadInto(fd, &buf, errmsg);
+    /* Read the checksums. */
+    rv = uvConsumeContent(content, offset, sizeof(uint32_t) * 2, &checksums,
+                          errmsg);
     if (rv != 0) {
         ErrMsgTransfer(errmsg, uv->io->errmsg, "read preamble");
         return RAFT_IOERR;
     }
 
-    n = (size_t)byteFlip64(preamble[1]);
+    /* Read the first 8 bytes of the batch, which contains the number of entries
+     * in the batch. */
+    rv = uvConsumeContent(content, offset, sizeof(uint64_t), &batch, errmsg);
+    if (rv != 0) {
+        ErrMsgTransfer(errmsg, uv->io->errmsg, "read preamble");
+        return RAFT_IOERR;
+    }
+
+    n = (size_t)byteFlip64(*(uint64_t *)batch);
     if (n == 0) {
         ErrMsgPrintf(uv->io->errmsg, "entries count in preamble is zero");
         rv = RAFT_CORRUPT;
@@ -237,38 +261,33 @@ static int uvLoadEntriesBatch(struct uv *uv,
         goto err;
     }
 
-    /* Read the batch header, excluding the first 8 bytes containing the number
-     * of entries, which we have already read. */
+    /* Consume the batch header, excluding the first 8 bytes containing the
+     * number of entries, which we have already read. */
     header.len = uvSizeofBatchHeader(n);
-    header.base = raft_malloc(header.len);
-    if (header.base == NULL) {
-        rv = RAFT_NOMEM;
-        goto err;
-    }
-    *(uint64_t *)header.base = preamble[1];
+    header.base = batch;
 
-    buf.base = (uint8_t *)header.base + sizeof(uint64_t);
-    buf.len = header.len - sizeof(uint64_t);
-    rv = UvFsReadInto(fd, &buf, errmsg);
+    rv = uvConsumeContent(content, offset,
+                          uvSizeofBatchHeader(n) - sizeof(uint64_t), NULL,
+                          errmsg);
     if (rv != 0) {
         ErrMsgTransfer(errmsg, uv->io->errmsg, "read header");
         rv = RAFT_IOERR;
-        goto err_after_header_alloc;
+        goto err;
     }
 
     /* Check batch header integrity. */
-    crc1 = byteFlip32(*(uint32_t *)preamble);
+    crc1 = byteFlip32(((uint32_t *)checksums)[0]);
     crc2 = byteCrc32(header.base, header.len, 0);
     if (crc1 != crc2) {
         ErrMsgPrintf(uv->io->errmsg, "header checksum mismatch");
         rv = RAFT_CORRUPT;
-        goto err_after_header_alloc;
+        goto err;
     }
 
     /* Decode the batch header, allocating the entries array. */
     rv = uvDecodeBatchHeader(header.base, entries, n_entries);
     if (rv != 0) {
-        goto err_after_header_alloc;
+        goto err;
     }
 
     /* Calculate the total size of the batch data */
@@ -276,47 +295,39 @@ static int uvLoadEntriesBatch(struct uv *uv,
     for (i = 0; i < n; i++) {
         data.len += (*entries)[i].buf.len;
     }
+    data.base = (uint8_t *)content->base + *offset;
 
-    /* Read the batch data */
-    data.base = raft_malloc(data.len);
-    if (data.base == NULL) {
-        rv = RAFT_NOMEM;
-        goto err_after_header_decode;
-    }
-    rv = UvFsReadInto(fd, &data, errmsg);
+    /* Consume the batch data */
+    rv = uvConsumeContent(content, offset, data.len, NULL, errmsg);
     if (rv != 0) {
         ErrMsgTransfer(errmsg, uv->io->errmsg, "read data");
         rv = RAFT_IOERR;
-        goto err_after_data_alloc;
+        goto err_after_header_decode;
     }
 
     /* Check batch data integrity. */
-    crc1 = byteFlip32(*((uint32_t *)preamble + 1));
+    crc1 = byteFlip32(((uint32_t *)checksums)[1]);
     crc2 = byteCrc32(data.base, data.len, 0);
     if (crc1 != crc2) {
         ErrMsgPrintf(uv->io->errmsg, "data checksum mismatch");
         rv = RAFT_CORRUPT;
-        goto err_after_data_alloc;
+        goto err_after_header_decode;
     }
 
-    uvDecodeEntriesBatch(&data, *entries, *n_entries);
+    uvDecodeEntriesBatch(content->base, *offset - data.len, *entries,
+                         *n_entries);
 
-    raft_free(header.base);
-
-    *last = UvFsIsAtEof(fd);
+    *last = *offset == content->len;
 
     return 0;
 
-err_after_data_alloc:
-    raft_free(data.base);
 err_after_header_decode:
-    raft_free(*entries);
-err_after_header_alloc:
-    raft_free(header.base);
+    HeapFree(*entries);
 err:
     *entries = NULL;
     *n_entries = 0;
     assert(rv != 0);
+    *offset = start;
     return rv;
 }
 
@@ -351,10 +362,11 @@ int uvSegmentLoadClosed(struct uv *uv,
                         size_t *n)
 {
     bool empty;                     /* Whether the file is empty */
-    uv_file fd;                     /* Segment file descriptor */
     uint64_t format;                /* Format version */
     bool last;                      /* Whether the last batch was reached */
     struct raft_entry *tmp_entries; /* Entries in current batch */
+    struct raft_buffer buf;         /* Segment file content */
+    size_t offset;                  /* Content read cursor */
     unsigned tmp_n;                 /* Number of entries in current batch */
     unsigned expected_n; /* Number of entries that we expect to find */
     int i;
@@ -377,14 +389,14 @@ int uvSegmentLoadClosed(struct uv *uv,
     }
 
     /* Open the segment file. */
-    rv = uvOpenSegmentFile(uv, info->filename, &fd, &format);
+    rv = uvReadSegmentFile(uv, info->filename, &buf, &format);
     if (rv != 0) {
         goto err;
     }
     if (format != UV__DISK_FORMAT) {
         ErrMsgPrintf(uv->io->errmsg, "unexpected format version %ju", format);
         rv = RAFT_CORRUPT;
-        goto err_after_open;
+        goto err_after_read;
     }
 
     /* Load all batches in the segment. */
@@ -392,13 +404,13 @@ int uvSegmentLoadClosed(struct uv *uv,
     *n = 0;
 
     last = false;
+    offset = sizeof format;
     for (i = 1; !last; i++) {
-        off_t offset;
-        rv = uvLoadEntriesBatch(uv, fd, &tmp_entries, &tmp_n, &offset, &last);
+        rv = uvLoadEntriesBatch(uv, &buf, &tmp_entries, &tmp_n, &offset, &last);
         if (rv != 0) {
-            ErrMsgWrapf(uv->io->errmsg, "entries batch %u starting at byte %ju",
+            ErrMsgWrapf(uv->io->errmsg, "entries batch %u starting at byte %zu",
                         i, offset);
-            goto err_after_open;
+            goto err_after_read;
         }
         rv = extendEntries(tmp_entries, tmp_n, entries, n);
         if (rv != 0) {
@@ -417,8 +429,6 @@ int uvSegmentLoadClosed(struct uv *uv,
     assert(i > 1);  /* At least one batch was loaded. */
     assert(*n > 0); /* At least one entry was loaded. */
 
-    close(fd);
-
     return 0;
 
 err_after_batch_load:
@@ -427,11 +437,11 @@ err_after_batch_load:
 
 err_after_extend_entries:
     if (*entries != NULL) {
-        entryBatchesDestroy(*entries, *n);
+        HeapFree(*entries);
     }
 
-err_after_open:
-    close(fd);
+err_after_read:
+    HeapFree(buf.base);
 
 err:
     assert(rv != 0);
@@ -439,6 +449,22 @@ err:
     return rv;
 }
 
+/* Check if the content of the segment file contains all zeros from the current
+ * offset onward. */
+static bool uvContentHasOnlyTrailingZeros(const struct raft_buffer *buf,
+                                          size_t offset)
+{
+    size_t i;
+
+    for (i = offset; i < buf->len; i++) {
+        if (((char *)buf->base)[i] != 0) {
+            return false;
+        }
+    }
+
+    return true;
+}
+
 /* Load all entries contained in an open segment. */
 static int uvLoadOpenSegment(struct uv *uv,
                              struct uvSegmentInfo *info,
@@ -451,11 +477,11 @@ static int uvLoadOpenSegment(struct uv *uv,
     bool empty;                     /* Whether the segment file is empty */
     bool remove = false;            /* Whether to remove this segment */
     bool last = false;              /* Whether the last batch was reached */
-    uv_file fd;                     /* Segment file descriptor */
     uint64_t format;                /* Format version */
     size_t n_batches = 0;           /* Number of loaded batches */
     struct raft_entry *tmp_entries; /* Entries in current batch */
-    off_t offset;                   /* Offset of last batch processed */
+    struct raft_buffer buf = {0};   /* Segment file content */
+    size_t offset;                  /* Content read cursor */
     unsigned tmp_n_entries;         /* Number of entries in current batch */
     int i;
     char errmsg[RAFT_ERRMSG_BUF_SIZE];
@@ -477,69 +503,55 @@ static int uvLoadOpenSegment(struct uv *uv,
         goto done;
     }
 
-    rv = uvOpenSegmentFile(uv, info->filename, &fd, &format);
+    rv = uvReadSegmentFile(uv, info->filename, &buf, &format);
     if (rv != 0) {
         goto err;
     }
 
     /* Check that the format is the expected one, or perhaps 0, indicating that
      * the segment was allocated but never written. */
+    offset = sizeof format;
     if (format != UV__DISK_FORMAT) {
         if (format == 0) {
-            rv = UvFsFileHasOnlyTrailingZeros(fd, &all_zeros, errmsg);
-            if (rv != 0) {
-                tracef("check if %s is zeroed: %s", info->filename, errmsg);
-                rv = RAFT_IOERR;
-                goto err_after_open;
-            }
+            all_zeros = uvContentHasOnlyTrailingZeros(&buf, offset);
             if (all_zeros) {
                 /* This is equivalent to the empty case, let's remove the
                  * segment. */
                 tracef("remove zeroed open segment %s", info->filename);
                 remove = true;
-                close(fd);
+                HeapFree(buf.base);
+                buf.base = NULL;
                 goto done;
             }
         }
         ErrMsgPrintf(uv->io->errmsg, "unexpected format version %ju", format);
         rv = RAFT_CORRUPT;
-        goto err_after_open;
+        goto err_after_read;
     }
 
     /* Load all batches in the segment. */
     for (i = 1; !last; i++) {
-        rv = uvLoadEntriesBatch(uv, fd, &tmp_entries, &tmp_n_entries, &offset,
+        rv = uvLoadEntriesBatch(uv, &buf, &tmp_entries, &tmp_n_entries, &offset,
                                 &last);
         if (rv != 0) {
-            int rv2;
-
             /* If this isn't a decoding error, just bail out. */
             if (rv != RAFT_CORRUPT) {
                 ErrMsgWrapf(uv->io->errmsg,
-                            "entries batch %u starting at byte %ju", i, offset);
-                goto err_after_open;
+                            "entries batch %u starting at byte %zu", i, offset);
+                goto err_after_read;
             }
 
             /* If this is a decoding error, and not an OS error, check if the
              * rest of the file is filled with zeros. In that case we assume
              * that the server shutdown uncleanly and we just truncate this
              * incomplete data. */
-            lseek(fd, offset, SEEK_SET);
-
-            rv2 = UvFsFileHasOnlyTrailingZeros(fd, &all_zeros, uv->io->errmsg);
-            if (rv2 != 0) {
-                ErrMsgWrapf(uv->io->errmsg, "check if %s is zeroed",
-                            info->filename);
-                rv = RAFT_IOERR;
-                goto err_after_open;
-            }
-
+            all_zeros = uvContentHasOnlyTrailingZeros(&buf, offset);
             if (!all_zeros) {
                 tracef("%s has non-zero trail", info->filename);
             }
 
             Tracef(uv->tracer,
-                   "truncate open segment %s at %jd (batch %d), since it has "
+                   "truncate open segment %s at %zu (batch %d), since it has "
                    "corrupted "
                    "entries",
                    info->filename, offset, i);
@@ -558,10 +570,9 @@ static int uvLoadOpenSegment(struct uv *uv,
         *next_index += tmp_n_entries;
     }
 
-    rv = close(fd);
-    assert(rv == 0);
-
     if (n_batches == 0) {
+        HeapFree(buf.base);
+        buf.base = NULL;
         remove = true;
     }
 
@@ -573,15 +584,20 @@ done:
         if (rv != 0) {
             tracef("unlink %s: %s", info->filename, errmsg);
             rv = RAFT_IOERR;
-            goto err_after_open;
+            goto err_after_read;
         }
     } else {
-        char filename[UV__FILENAME_LEN];
+        char filename[UV__SEGMENT_FILENAME_BUF_SIZE];
         raft_index end_index = *next_index - 1;
 
         /* At least one entry was loaded */
         assert(end_index >= first_index);
-        sprintf(filename, UV__CLOSED_TEMPLATE, first_index, end_index);
+        int nb = snprintf(filename, sizeof(filename), UV__CLOSED_TEMPLATE, first_index, end_index);
+        if ((nb < 0) || ((size_t)nb >= sizeof(filename))) {
+            tracef("snprintf failed: %d", nb);
+            rv = RAFT_IOERR;
+            goto err;
+        }
 
         tracef("finalize %s into %s", info->filename, filename);
 
@@ -596,7 +612,12 @@ done:
         info->is_open = false;
         info->first_index = first_index;
         info->end_index = end_index;
-        strcpy(info->filename, filename);
+        memset(info->filename, '\0', sizeof(info->filename));
+        _Static_assert(sizeof(info->filename) >= sizeof(filename), "Destination buffer too small");
+        /* info->filename is zeroed out, info->filename is at least as large as
+         * filename and we checked that nb < sizeof(filename) -> we won't
+         * overflow and the result will be zero terminated. */
+        memcpy(info->filename, filename, (size_t)nb);
     }
 
     return 0;
@@ -605,8 +626,10 @@ err_after_batch_load:
     raft_free(tmp_entries[0].batch);
     raft_free(tmp_entries);
 
-err_after_open:
-    close(fd);
+err_after_read:
+    if (buf.base != NULL) {
+        HeapFree(buf.base);
+    }
 
 err:
     assert(rv != 0);
@@ -727,7 +750,7 @@ int uvSegmentBufferAppend(struct uvSegmentBuffer *b,
     crc2 = 0;
     for (i = 0; i < n_entries; i++) {
         const struct raft_entry *entry = &entries[i];
-        /* TODO: enforce the requirment of 8-byte aligment also in the
+        /* TODO: enforce the requirement of 8-byte alignment also in the
          * higher-level APIs. */
         assert(entry->buf.len % sizeof(uint64_t) == 0);
         memcpy(cursor, entry->buf.base, entry->buf.len);
